# This script makes assumptions about the existing sections and option names
# in the associated config.ini file.  If the provided config.ini file is directly
# edited, then this file should be reviewed for breaking changes.
# For the relative filepaths to work, this script must be run from the outer porthole directory.


from flask import Flask, render_template, flash
from flask_wtf import FlaskForm
from wtforms import StringField, SubmitField, PasswordField, RadioField, SelectField
from wtforms.validators import InputRequired
from configparser import ConfigParser
from flask_bootstrap import Bootstrap
from .connections import ConnectionManager
# from . import gui_app

gui_app = Flask('porthole')
gui_app.config['SECRET_KEY'] = 'flarp'

#flask_bootstrap fills in some boilerplate in the template
Bootstrap(gui_app)

#ConfigParser helps with manipulating config files
parser = ConfigParser()
config_file = 'config/config.ini'


# This function takes a dictionary and updates the config file in the base directory.
# This function checks for specific keys and only updates when key exists and value is not an empty string.
def edit_config(dict):
    parser.read(config_file)
    if 'default_base_file_path' in dict and dict['default_base_file_path'] != '':
        parser.set('Default', 'base_file_path', dict['default_base_file_path'])
    if 'default_query_path' in dict and dict['default_query_path'] != '':
        parser.set('Default', 'query_path', dict['default_query_path'])
    if 'default_database' in dict and dict['default_database'] != '':
        parser.set('Default', 'database', dict['default_database'])
    if 'default_notification_recipient' in dict and dict['default_notification_recipient'] != '':
        parser.set('Default', 'notification_recipient', dict['default_notification_recipient'])
    if 'email_username' in dict and dict['email_username'] != '':
        parser.set('Email', 'username', dict['email_username'])
    if 'email_password' in dict and dict['email_password'] != '':
        parser.set('Email', 'password', dict['email_password'])
    if 'email_host' in dict and dict['email_host'] != '':
        parser.set('Email', 'host', dict['email_host'])
    if 'email_disabled' in dict and str(dict['email_disabled']) != '':
        parser.set('Email', 'disabled', str(dict['email_disabled']))
    if 'email_signature' in dict and dict['email_signature'] != '':
        parser.set('Email', 'signature', dict['email_signature'])
    if 'logging_server' in dict and dict['logging_server'] != '':
        parser.set('Logging', 'server', dict['logging_server'])
    if 'logging_to_file' in dict and dict['logging_to_file'] != '':
        parser.set('Logging', 'log_to_file', dict['logging_to_file'])
    if 'logging_file' in dict and dict['logging_file'] != '':
        parser.set('Logging', 'logfile', dict['logging_file'])
    if 'logging_db' in dict and dict['logging_db'] != '':
        parser.set('Logging', 'db', dict['logging_db'])
    if 'debug_mode' in dict and dict['debug_mode'] != '':
        parser.set('Debug', 'debug_mode', dict['debug_mode'])
    if 'debug_recipients' in dict and dict['debug_recipients'] != '':
        parser.set('Debug', 'debug_recipients', dict['debug_recipients'])
    if 'admin_email' in dict and dict['admin_email'] != '':
        parser.set('Admin', 'admin_email', dict['admin_email'])
    with open(config_file, 'w') as configfile:
        parser.write(configfile)

# This function takes a dictionary and updates the config file in the base directory.
# This function will add a new section to the config file if the provided section name (connection_name)
# is unique.  Otherwise it will update the section with the options provided.  If the section is new
# the connection_name will be appended to the connections option in the Default section (this assists
# in iteration over the connection sections).  Blank, extra, and missing options are ignored.
def add_connection(dict):
    parser.read(config_file)
    if 'connection_name' in dict and dict['connection_name'] != '':
        connection_name = dict['connection_name']
        connections = parser.get('Default', 'connections')
        if connection_name not in connections.split(", "):
            parser.add_section(connection_name)
            parser.set('Default', 'connections', connections + ', ' + connection_name)
    if 'rdbms' in dict and dict['rdbms'] != '':
        parser.set(connection_name, 'rdbms', dict['rdbms'])
    if 'connection_host' in dict and dict['connection_host'] != '':
        parser.set(connection_name, 'host', dict['connection_host'])
    if 'connection_port' in dict and dict['connection_port'] != '':
        parser.set(connection_name, 'port', dict['connection_port'])
    if 'connection_user' in dict and dict['connection_user'] != '':
        parser.set(connection_name, 'user', dict['connection_user'])
    if 'connection_password' in dict and dict['connection_password'] != '':
        parser.set(connection_name, 'password', dict['connection_password'])
    if 'schema' in dict and dict['schema'] != '':
        parser.set(connection_name, 'schema', dict['schema'])
    with open(config_file, 'w') as configfile:
        parser.write(configfile)

# This function takes a string and updates the config file in the base directory.
# If a section name matches the provided string, then that section and all its options are removed.
# It is removed from the connections option in the Default section, if present.
def delete_connection(connection_name):
    print(connection_name)
    parser.read(config_file)
    connections = parser.get('Default', 'connections')
    connections_list = connections.split(", ")
    if connection_name in connections_list:
        parser.remove_section(connection_name)
        connections_list.remove(connection_name)
        parser.set('Default', 'connections',', '.join(connections_list))
    with open(config_file, 'w') as configfile:
        parser.write(configfile)

# This function takes no arguments and returns a dictionary of dictionaries.
# The outer dictionary has each section in the config file as keys. The inner dictionary
# (i.e. the values of the outer dictionary keys) contains a dictionary where the option names
# are the keys, and the option values are the values.
# e.g. to get the query_path in the Default section: all_config_options['Default']['query_path']
def read_config():
    parser.read(config_file)
    all_config_options = {}
    sections = parser.sections()
    for section in sections:
        all_config_options.update({section : {}})
        options = parser.options(section)
        for option in options:
            all_config_options[section].update({option : parser.get(section, option)})
    return all_config_options

# This class defines the form used to add and edit connection sections to the config file.
class ConnectionForm(FlaskForm):
    connection_name = StringField('connection_name')
    rdbms = StringField('rdbms')
    connection_host = StringField('connection_host')
    connection_port = StringField('connection_port')
    connection_user = StringField('connection_user')
    connection_password = StringField('connection_password')
    schema = StringField('schema')
    delete_connection = StringField('Delete?')
    connection_submit = SubmitField('Save Connection')

# This class defines the form used to edit config sections other than connections.
class ConfigForm(FlaskForm):
    default_base_file_path = StringField('default_base_file_path')#, validators=[InputRequired()])
    default_query_path = StringField('default_query_path')
    default_database = SelectField('default_database')
    default_notification_recipient = StringField('default_notification_recipient')
    email_username = StringField('email_username')
    email_password = PasswordField('email_password')
    email_host = StringField('email_host')
    email_disabled = RadioField('email_disabled', choices=[('True','True'),('False','False')])
    email_signature = StringField('email_signature')
    logging_server = SelectField('logging_server')
    logging_to_file = RadioField('Log to File', choices=[('True', 'True'),('False', 'False')])
    logging_file = StringField('Log File Name')
    logging_db = StringField('logging_db')
    debug_mode = RadioField('debug_mode', choices=[('True', 'True'),('False', 'False')])
    debug_recipients = StringField('debug_recipients')
    admin_email = StringField('admin_email')
    config_submit = SubmitField('Save Settings')

# This is the endpoint for displaying and editing the application settings.
# All of the config sections and options are passed to the front end, along with options for
# select inputs in the rendered form.
# This endpoint renders two forms.  If either form is submitted and validated, it will update
# the config file and refresh the page showing the updated values.
@gui_app.route('/config', methods=['GET', 'POST'])
def config():
    all_config_options=read_config()
    connections = all_config_options["Default"]["connections"].split(", ")
    config_form = ConfigForm()
    connection_form = ConnectionForm()
    connection_choices = [(c,c) for c in connections]
    connection_choice_additions = [('None Selected','None Selected')]
    print(connection_choices + connection_choice_additions)
    config_form.default_database.choices = connection_choices + connection_choice_additions
    config_form.logging_server.choices = connection_choices + connection_choice_additions
    rdbms_options = ['MySQL', 'SQLite']
    rdbms_choices = [(i,i) for i in rdbms_options]
    connection_form.rdbms.choices = rdbms_choices

    if config_form.config_submit.data and config_form.validate():
        edit_config(config_form.data)
        flash('Settings Updated')
        all_config_options=read_config()
        connections = all_config_options["Default"]["connections"].split(", ")
        print(config_form.data)
        return render_template("settings.html", config_form=config_form
                                            , connection_form=connection_form
                                            , all_config_options=all_config_options
                                            , connections=connections
                                            , rdbms_options=rdbms_options)

    if connection_form.connection_submit.data and connection_form.validate():
        if connection_form.data['delete_connection'] == 'True':
            delete_connection(connection_form.data['connection_name'])
        else:
            add_connection(connection_form.data)
        print(connection_form.data)
        all_config_options=read_config()
        connections = all_config_options["Default"]["connections"].split(", ")
        return render_template("settings.html", config_form=config_form
                                            , connection_form=connection_form
                                            , all_config_options=all_config_options
                                            , connections=connections
                                            , rdbms_options=rdbms_options)

    return render_template("settings.html", config_form=config_form
                                        , connection_form=connection_form
                                        , all_config_options=all_config_options
                                        , connections=connections
                                        , rdbms_options=rdbms_options)

@gui_app.route('/api/test_connection/<connection_name>', methods=['GET', 'POST'])
def test_connection(connection_name):
    conn = ConnectionManager(db = connection_name)
    try:
        conn.connect()
        conn.close()
        return 'Connected Successfully'
    except:
        return 'Failed to Connect'


if __name__ == '__main__':
    gui_app.run(debug=False)



"""
Created on 03/23/2017.

Contains various helper functions related to time. By default,
dates and timestamps are returned as formatted strings. Just pass
string=False to change this behavior.

Sample Usage:

>>> from helpers import TimeHelper
>>> TimeHelper.current_timestamp()
'2017-03-23 22:50:31'
>>> TimeHelper.today()
'2017-03-23'
>>> TimeHelper.yesterday()
'2017-03-22'
>>> TimeHelper.today(string=False)
datetime.date(2017, 3, 23)

"""

from datetime import datetime, timedelta


def current_timestamp(string=True):
    timestamp = datetime.utcnow()
    if string:
        return timestamp_to_string(timestamp)
    else:
        return timestamp

def now(string=True):
    return current_timestamp(string)

def today(string=True):
    date = datetime.utcnow().date()
    if string:
        return date_to_string(date)
    else:
        return date

def yesterday(string=True):
    date = today(False) - timedelta(days=1)
    if string:
        return date_to_string(date)
    else:
        return date

def timestamp_to_string(timestamp):
    return timestamp.strftime('%Y-%m-%d %-H:%M:%S')

def date_to_string(date):
    return date.strftime('%Y-%m-%d')


if __name__ == '__main__':
    pass



from .app import config
from .connections import ConnectionPool
from .mailer import Mailer
from .components import (DatabaseLogger,
                         Loggable,
                         RecipientsChecker,
                         ReportActiveChecker,
                         ReportErrorNotifier,
                         ReportWriter)
from .logger import PortholeLogger

logger = PortholeLogger(name=__name__)


class BasicReport(Loggable):
    """
    A basic report with bare minimum functionality.

    Keyword arguments
    :report_title: Used in the filename of the resulting report.

    """
    def __init__(self, report_title):
        # -----------------------------------------
        # Assign arguments to instance attributes.
        # -----------------------------------------
        self.report_title = report_title
        self.attachments = []
        self.report_writer = None
        self.error_log = []
        self.to_recipients = []
        self.cc_recipients = []
        self.email_sent = False
        self.failure_notification_sent = False
        self.file_path = config['Default'].get('base_file_path')
        self.default_db = config['Default'].get('database')
        self.conns = ConnectionPool([self.default_db])

    def __del__(self):
        try:
            self.conns.close_all()
        except:
            pass

    def get_conn(self, db):
        return self.conns.pool.get(db)

    def add_conn(self, db):
        return self.conns.add_connection(db)

    def build_file(self):
        report_writer = ReportWriter(report_title=self.report_title, log_to=self.error_log)
        report_writer.build_file()
        self.attachments.append(report_writer.report_file)
        self.report_writer = report_writer

    def create_worksheet_from_query(self,
                                    sheet_name,
                                    db=None,
                                    query={},
                                    sql=None,
                                    **kwargs):
        """Delegates functionality to ReportWriter."""
        if db is None:
            db = self.default_db
        cm = self.add_conn(db)
        self.report_writer.create_worksheet_from_query(cm=cm,
                                                        sheet_name=sheet_name,
                                                        query=query,
                                                        sql=sql,
                                                        **kwargs)

    def make_worksheet(self, sheet_name, query_results, **kwargs):
        """Delegates functionality to ReportWriter."""
        self.report_writer.make_worksheet(sheet_name=sheet_name,
                                            query_results=query_results,
                                            **kwargs
                                            )

    def build_email(self):
        """Instantiates Mailer object using provided parameters."""
        email = Mailer()
        email.recipients = self.to_recipients
        email.cc_recipients = self.cc_recipients
        email.subject = self.subject
        email.message = self.message
        email.attachments = self.attachments
        self.email = email

    def send_email(self):
        """Executes the send_email method of the Mailer."""
        try:
            self.email.send_email()
            self.email_sent = True
        except Exception as e:
            logger.exception(e)
            self.log_error("Unable to send email")

    def check_whether_to_send_email(self):
        """Determines whether email should be sent based given errors and settings."""
        if self.error_log:
            return False
        else:
            return True

    def build_and_send_email(self):
        """If email should be sent, builds email object and sends email."""
        if self.check_whether_to_send_email():
            self.build_email()
            self.send_email()

    def execute(self):
        """
        After instantiation and setup, and after
        any query results have been added to the workbook,
        this method executes and finalizes the report,
        ensures errors are handled, and performs cleanup.
        """
        if self.report_writer:
            self.report_writer.close_workbook()
        self.build_and_send_email()
        if self.error_log:
            self.send_failure_notification()
        self.conns.close_all()

    def send_failure_notification(self):
        notifier = ReportErrorNotifier( report_title=self.report_title,
                                        error_log=self.error_log)
        notifier.send_log_by_email()
        if notifier.notified:
            self.failure_notification_sent = True


class GenericReport(BasicReport, Loggable):
    """A generic report object to be used to facilitate
    easy setup and configuration of new automated reports.

    Keyword arguments
    :report_title: Used in the filename of the resulting report.
    :report_name: Uniquely identifies a report. Used for recipient lookup
        and logging.
    :logging_enabled: (Optional) Default value of True. If logging is enabled,
        the report will attempt to create a log record each time it is instantiated.
        Only disable if the report needs to be run without connection to the standard
        reporting database.

    Here is an example of sample usage:

    report = GenericReport(
                            report_name='test_report',
                            report_title='Test Report'
                            )
    report.build_file()
    report.create_worksheet_from_query(
                                        sheet_name='Sheet1',
                                        query={'filename': 'test_query'}
                                        )
    report.subject = "Today's Test Report"
    report.message = "Attached is the latest report"
    report.execute()

    Notes on sample code:
    (1) A report is not required to have an attachment,
        but an attachment is assumed.
    (2) An Excel attachment can contain any number
        of worksheets. Simply call create_worksheet_from_query
        with the appropriate query parameters and worksheet
        name, as many times as you like. Worksheet names
        must be valid and unique.
    (3) After providing a subject and message body,
        the execute method will take care of the rest,
        including email circulation, error handling,
        and notification.

    Notes on database configuration:
    (1) Reports must have a record in the automated_reports table,
        including report_name and active flag (1 or 0).
    (2) Reports also must have at least one recipient
        in the automated_report_recipients table, and at least
        one recipient must be designed with a recipient type of
        'to'. Each report can have one or more recipients with
        a recipient type of 'cc'.
    (3) Report recipient records are associated with:
        (a) automated_reports through report_id
        (b) automated_report_contacts through contact_id
    (4) If you need a recipient who is not in the
        automated_report_contacts table, add them to it.

    """
    def __init__(self, report_title, report_name, logging_enabled=True, send_if_blank=True):
        # -----------------------------------------
        # Assign arguments to instance attributes.
        # -----------------------------------------
        super().__init__(report_title=report_title)
        self.report_name = report_name
        self.logging_enabled = logging_enabled
        self.send_if_blank = send_if_blank
        self.check_if_active()
        if self.active:
            self.initialize_db_logger()

    @property
    def record_count(self):
        if self.report_writer:
            return self.report_writer.record_count
        else:
            return 0

    def initialize_db_logger(self):
        self.db_logger = DatabaseLogger( cm=self.get_conn(self.default_db),
                                         report_name=self.report_name,
                                         log_to=self.error_log)
        self.db_logger.create_record()

    def check_if_active(self):
        self.active = ReportActiveChecker(  cm=self.get_conn(self.default_db),
                                            report_name=self.report_name,
                                            log_to=self.error_log)

    def get_recipients(self):
        """Performs lookup in database for report recipients based on report name."""

        checker = RecipientsChecker( cm=self.get_conn(self.default_db),
                                     report_name=self.report_name,
                                     log_to=self.error_log)
        self.to_recipients, self.cc_recipients = checker.get_recipients()

    def check_whether_to_send_email(self):
        """Determines whether email should be sent based given errors, settings, and result count."""
        if self.error_log:
            return False
        elif not self.send_if_blank and self.record_count == 0:
            return False
        else:
            return True

    def execute(self):
        """
        After instantiation and setup, and after
        any query results have been added to the workbook,
        this method executes and finalizes the report,
        ensures errors are handled, and performs cleanup.
        """
        if self.report_writer:
            self.report_writer.close_workbook()
        if self.active:
            self.build_and_send_email()
            self.db_logger.finalize_record()
        if self.error_log:
            self.send_failure_notification()
        self.conns.close_all()




if __name__ == '__main__':
    pass



"""
Mailer.py

"""

import sys
import os
import smtplib
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from .app import config
from .logger import PortholeLogger

logger = PortholeLogger(name=__name__)

COMMASPACE = ', '


class Mailer:
    def __init__(self, recipients=[], cc_recipients=[], **kwargs):
        self.recipients = recipients
        self.cc_recipients = cc_recipients
        self.properties = kwargs
        self.debug_mode = config.getboolean('Debug', 'debug_mode')
        debug_recipients = config['Debug']['debug_recipients']
        self.debug_recipients = [recip.strip() for recip in debug_recipients.split(';')]
        self.username = config['Email']['username']
        self.send_from = config['Email'].get('send_from', self.username)
        self.password = config['Email']['password']
        self.host = config['Email'].get('host', 'smtp.office365.com')
        self.signature = config['Email']['signature']
        self.admin_email = config['Admin']['admin_email']
        self.attachments = None
        self.all_sent_to_recipients = None

    def send_email(self):
        if config['Email']['disabled'] == 'True':
            logger.info("Email disabled - exiting mailer.")
            exit()

        # Create the enclosing (outer) message
        outer = MIMEMultipart('alternative')
        outer['Subject'] = self.subject
        outer['From'] = self.send_from
        if self.debug_mode:
            outer['To'] = COMMASPACE.join(self.debug_recipients)
        else:
            outer['To'] = COMMASPACE.join(self.recipients)
            outer['CC'] = COMMASPACE.join(self.cc_recipients)

        msg = MIMEBase('application', "octet-stream")

        # Add the text of the email
        email_body = MIMEText(self.message, 'plain')
        outer.attach(email_body)

        # Add the attachments
        if self.attachments:
            for file in self.attachments:
                try:
                    with open(file, 'rb') as fp:
                        msg.set_payload(fp.read())
                    encoders.encode_base64(msg)
                    msg.add_header('Content-Disposition',
                                   'attachment',
                                   filename=os.path.basename(file))
                    outer.attach(msg)
                except Exception as e:
                    logger.exception(e)
                    raise

        composed = outer.as_string()

        if self.debug_mode:
            logger.info("""Debug mode active. Sending only to designated debug recipient(s).

Active report would have been sent to:
            "To" Recipients: {}
            "CC" Recipients: {}""".format(self.recipients, self.cc_recipients))
            all_recipients = self.debug_recipients
        else:
            all_recipients = set(self.recipients + self.cc_recipients)
            all_recipients = list(filter(None, all_recipients))
            all_recipients.append(self.admin_email)

        try:
            with smtplib.SMTP(host=self.host, port='587') as s:
                s.ehlo()
                s.starttls()
                s.login(self.username, self.password)
                s.sendmail(self.send_from,
                           all_recipients,
                           composed)
                self.all_sent_to_recipients = all_recipients
                s.close()
                logger.info("Email with subject '{}...' sent to {} recipients".format(self.subject, len(all_recipients)))
        except Exception as e:
            logger.exception(e)
            raise

if __name__ == '__main__':
    pass



from .queries import QueryResult


class ResultFilter(object):
    """
    Used for splitting a QueryResult on a specified attribute into sub-results.
    Subsets of the provided QueryResultw ill be created for each distinct value
    of the specified field.
    ResultFilter objects are iterables which iterate through the filtered_results
    attribute.

    Keyword arguments
    :result_to_filter: (QueryResult) The object to be split.
    :filter_by: (str) The field name on which to split.

    """
    def __init__(self, result_to_filter, filter_by):
        self.result_to_filter = result_to_filter
        self.headers = result_to_filter.field_names
        self.data = result_to_filter.result_data
        self.filter_by = filter_by
        self.keys = []
        self.filtered_data = {}
        self.filtered_results = {}
        self.filter_by = filter_by
        if filter_by not in self.headers:
            raise ValueError("Provided headers must contain filter_by value.")

    def filter(self):
        for row in self.data:
            key = row[self.filter_by]
            if key not in self.keys:
                self.keys.append(key)
                self.filtered_data[key] = []
            self.filtered_data[key].append(list(row))
        for key in self.keys:
            self.filtered_results[key] = QueryResult(
                field_names=list(self.headers),
                result_data=self.filtered_data[key]
                )

    def __iter__(self):
        self.pos = 0
        self.end = len(self.keys) - 1
        return self

    def __next__(self):
        if self.pos <= self.end:
            self.pos += 1
            key = self.keys[self.pos - 1]
            return key, self.filtered_results[key]
        else:
            raise StopIteration
